package benchmark;

import java.io.File;
import java.io.IOException;
import java.rmi.UnexpectedException;
import java.util.Arrays;
import java.util.Comparator;

import solver.approximationHClustPivot.HcPivot;
import solver.lpBased.HierClusterLpSolver;
import benchmark.helpClasses.BenchResult;
import benchmark.helpClasses.LogFileWriter;
import benchmark.helpClasses.timeBoundedApplications.FindOptimalSolutionWithinTime;
import dataReductionRules.UnsolvableException;
import distanceFunction.SymmetricDiscreteDistanceFunction;

public class RealWorldDataBenchTestLP
{
	public static String Aprox_Equal_LP = "ApELp";

	static final int maxNumberSeconds = 60 * 5;

	public static String OPTIMAL_SOLVED_BY_LP = "optLP";

	public static String STATUS = "stStatus";

	private static int extractNumber(final String str)
	{
		final int beginIndex = str.indexOf("_n_") + 3;
		final int endIndex = str.indexOf("_", beginIndex);
		final String numberSTR = str.substring(beginIndex, endIndex);
		final int i = Integer.parseInt(numberSTR.equalsIgnoreCase("") ? "0" : numberSTR);
		return i;
	}

	public static void main(final String[] args) throws IOException
	{
		final File instanceFolder = new File("./testInstances");
		for (final String folderName : instanceFolder.list())
		{
			final File folder = new File(instanceFolder.getCanonicalPath() + "/" + folderName);
			if (folder.isDirectory())
			{

				final LogFileWriter logWriter = new LogFileWriter("realWorldLog_" + folderName + ".csv", true, true);

				final String[] fileList = folder.list();
				class StringComparator implements Comparator<String>
				{

					@Override
					public int compare(final String o1, final String o2)
					{
						return RealWorldDataBenchTestLP.extractNumber(o1) - RealWorldDataBenchTestLP.extractNumber(o2);
					}

				}
				;
				Arrays.sort(fileList, new StringComparator());

				RealWorldDataBenchTestLP.writeHeadline(logWriter);

				for (final String file : fileList)
				{
					if (file.endsWith("dat"))
					{
						final SymmetricDiscreteDistanceFunction dist = SymmetricDiscreteDistanceFunction.createInstanceFromFile(instanceFolder
								.getCanonicalPath() + "/" + folderName + "/" + file);
						if (dist.getNoOfElements() <= 70)
						{
							RealWorldDataBenchTestLP.writeResults(RealWorldDataBenchTestLP.runTests(dist, false, true, false),
									file + "\t" + dist.getNoOfElements(), logWriter);
						}
					}
				}
				logWriter.close();
			}
		}
	}

	public static void reverse(final String[] b)
	{
		int left = 0; // index of leftmost element
		int right = b.length - 1; // index of rightmost element

		while (left < right)
		{
			// exchange the left and right elements
			final String temp = b[left];
			b[left] = b[right];
			b[right] = temp;

			// move the bounds toward the center
			left++;
			right--;
		}
	}

	public static BenchResult[] runTests(final SymmetricDiscreteDistanceFunction dist, final boolean runILP, final boolean runST, final boolean runLP)
			throws IOException, UnexpectedException
	{
		System.gc();

		BenchResult approxResult;

		int d_approx = -2;
		{// approximation algorithm
			final long begin = System.currentTimeMillis();
			final int repetitions = Math.min(10000, Math.max(100, dist.getNoOfElements() * dist.getNoOfElements()));
			d_approx = HcPivot.run_hclustpivot(dist, repetitions);
			approxResult = new BenchResult(d_approx, (System.currentTimeMillis() - begin) / 1000d);
		}

		// lp
		BenchResult lpResult = new BenchResult(-1, -1);
		BenchResult stBenchResult = new BenchResult(-1, -1);
		if (runLP)
		{
			double optLpSolution = -1;
			final long begin = System.currentTimeMillis();
			final HierClusterLpSolver lpSolver = HierClusterLpSolver.getStandardLpSolver(dist, -1, null);
			{
				try
				{
					lpSolver.solve();
					optLpSolution = lpSolver.getOptLpSolution();
				} catch (final UnsolvableException e)
				{

				}
				lpResult = new BenchResult(optLpSolution, (System.currentTimeMillis() - begin) / 1000d);
				if (lpSolver.isIlpSolution())
				{
					lpResult.setParameter(RealWorldDataBenchTestLP.OPTIMAL_SOLVED_BY_LP, 1);
				} else
				{
					lpResult.setParameter(RealWorldDataBenchTestLP.OPTIMAL_SOLVED_BY_LP, 0);
				}
			}

			if (Math.ceil(optLpSolution) == d_approx)
			{
				approxResult.setParameter(RealWorldDataBenchTestLP.Aprox_Equal_LP, 1);
			} else
			{
				approxResult.setParameter(RealWorldDataBenchTestLP.Aprox_Equal_LP, 0);
			}

			// searchTreeAlgorithm
			// if ( !lpSolver.isIlpSolution() )
			// {
			final FindOptimalSolutionWithinTime d = new FindOptimalSolutionWithinTime(RealWorldDataBenchTestLP.maxNumberSeconds, dist, true, true,
					true);
			d.runAlgorithm();
			stBenchResult = d.getBenchResult();
			if (d.isStopped())
			{
				stBenchResult.setParameter(RealWorldDataBenchTestLP.STATUS, "0");
			} else
			{
				stBenchResult.setParameter(RealWorldDataBenchTestLP.STATUS, "1");
			}
		}
		// }

		// ilp
		final BenchResult ilpResult;
		{
			if (runILP)
			{
				final long begin = System.currentTimeMillis();
				final HierClusterLpSolver ilpSolver = HierClusterLpSolver.getStandardLpSolver(dist, -1, null);
				ilpSolver.setILP(true);
				int ilpSol = -1;
				boolean solvable = true;
				try
				{
					ilpSolver.solve();
				} catch (final UnsolvableException e)
				{
					solvable = false;
				}
				if (solvable)
				{
					ilpSol = (int) ilpSolver.getOptLpSolution();
					if (!ilpSolver.isIlpSolution())
						throw new UnexpectedException("wrong ilp solution");
				} else
				{
					ilpSol = -1;
				}
				ilpResult = new BenchResult(ilpSol, (System.currentTimeMillis() - begin) / 1000d);
			} else
			{
				ilpResult = new BenchResult(-1, -1);
			}
		}

		BenchResult stBenchResultWitoutLP = new BenchResult(-1, -1);
		{
			if (runST)
			{
				final FindOptimalSolutionWithinTime d = new FindOptimalSolutionWithinTime(RealWorldDataBenchTestLP.maxNumberSeconds, dist, true,
						false, true);
				d.runAlgorithm();
				stBenchResultWitoutLP = d.getBenchResult();
			}

		}
		return new BenchResult[] { approxResult, lpResult, stBenchResult, ilpResult, stBenchResultWitoutLP };
	}

	public static void writeHeadline(final LogFileWriter logWriter) throws IOException
	{
		logWriter.write("name\tappK\tappTime\t" + RealWorldDataBenchTestLP.Aprox_Equal_LP + "\t"
				+ "lpK\tlpTime\tLpOpt\tstK\tstTime\tstIter\tstStatus\tilpK\tilpTime\tst_oLP_K\tst_oLP_time\tst_oLP_Iter\n");
	}

	public static void writeResults(final BenchResult approxResult, final BenchResult lpResult, final BenchResult stResult,
			final BenchResult ilpResult, final BenchResult stWithoutLP, final String name, final LogFileWriter logWriter) throws IOException
	{
		logWriter.write(name);
		// approx
		logWriter.write("\t" + SyntheticBenchmarks.timeFormatter.format(approxResult.getParameter()) + "\t"
				+ SyntheticBenchmarks.timeFormatter.format(approxResult.getTime()) + "\t"
				+ approxResult.getParameter(RealWorldDataBenchTestLP.Aprox_Equal_LP));

		// lp
		logWriter.write("\t" + SyntheticBenchmarks.timeFormatter.format(lpResult.getParameter()) + "\t"
				+ SyntheticBenchmarks.timeFormatter.format(lpResult.getTime()) + "\t"
				+ lpResult.getParameter(RealWorldDataBenchTestLP.OPTIMAL_SOLVED_BY_LP));

		// st
		logWriter.write("\t" + SyntheticBenchmarks.timeFormatter.format(stResult.getParameter()) + "\t"
				+ SyntheticBenchmarks.timeFormatter.format(stResult.getTime()) + "\t" + stResult.getNumberRecursiveSteps() + "\t"
				+ stResult.getParameter(RealWorldDataBenchTestLP.STATUS));

		// ilp
		logWriter.write("\t" + SyntheticBenchmarks.timeFormatter.format(ilpResult.getParameter()) + "\t"
				+ SyntheticBenchmarks.timeFormatter.format(ilpResult.getTime()));

		if (stWithoutLP != null)
		{
			logWriter.write("\t" + SyntheticBenchmarks.timeFormatter.format(stWithoutLP.getParameter()) + "\t"
					+ SyntheticBenchmarks.timeFormatter.format(stWithoutLP.getTime()) + "\t" + stWithoutLP.getNumberRecursiveSteps());
		}

		logWriter.write("\n");
		logWriter.flush();

	}

	public static void writeResults(final BenchResult approxResult, final BenchResult lpResult, final BenchResult stResult,
			final BenchResult ilpResult, final String name, final LogFileWriter logWriter) throws IOException
	{
		RealWorldDataBenchTestLP.writeResults(approxResult, lpResult, stResult, ilpResult, null, name, logWriter);
	}

	public static void writeResults(final BenchResult[] results, final String name, final LogFileWriter logWriter) throws IOException
	{
		final BenchResult approxResult = results[0];
		final BenchResult lpResult = results[1];
		final BenchResult stResult = results[2];
		final BenchResult ilpResult = results[3];
		final BenchResult noLPSTResult = results[4];
		RealWorldDataBenchTestLP.writeResults(approxResult, lpResult, stResult, ilpResult, noLPSTResult, name, logWriter);

	}
}