package benchmark;

import java.io.IOException;
import java.text.DecimalFormat;

import solver.approximationHClustPivot.hclustpivot;
import benchmark.helpClasses.BenchResult;
import benchmark.helpClasses.LogFileWriter;
import benchmark.helpClasses.TimeBoundedApplication;
import distanceFunction.SymmetricDiscreteDistanceFunction;
import distanceFunction.random.RandomDistanceFunctions;

public class SyntheticBenchmarks
{
	public static final int NO_ITERATIONS = 4;

	public static final DecimalFormat timeFormatter = new DecimalFormat("0.0##");

	public static BenchResult getBestApproxResult(final SymmetricDiscreteDistanceFunction dist, final int noRuns)
	{
		int minDist = Integer.MAX_VALUE;
		final long begin = System.currentTimeMillis();
		for (int i = 0; i < noRuns; i++)
		{ // run HClustPivot
			final SymmetricDiscreteDistanceFunction distance = new SymmetricDiscreteDistanceFunction(dist);
			final hclustpivot alg = new hclustpivot(distance);
			alg.solve();
			if (distance.isUltrametric())
			{
				minDist = Math.min(minDist, dist.getDistance(distance));
			}
		}
		return new BenchResult(minDist, (System.currentTimeMillis() - begin) / 1000d, -1);
	}

	private static String getMeanLogEntry(final double meanTime)
	{
		return meanTime == TimeBoundedApplication.STOP_TIME_VALUE ? "infin" : SyntheticBenchmarks.timeFormatter.format(meanTime);
	}

	private static int[][][] getUltrametrics(final int n, final int maxDistValue) throws IOException
	{
		final int[][][] ultrametricArr = new int[SyntheticBenchmarks.NO_ITERATIONS][][];
		// ultrametrics
		for (int i = 0; i < SyntheticBenchmarks.NO_ITERATIONS; i++)
		{
			final int maxSplit = Math.max(1, (int) Math.log(n));
			ultrametricArr[i] = RandomDistanceFunctions.getUltrametric(n, maxDistValue, maxSplit);
			// final String filePath = "testInstances/visu/" + "graphviz_" + i +
			// "_" + n + "_" +
			// maxDistValue + "_" + maxSplit
			// + ".data";
			// RandomDistanceFunctions.visualizeUltrametric(filePath, new
			// DistanceFunction(ultrametricArr[i]));
		}
		return ultrametricArr;
	}

	public static void main(final String[] args) throws IOException
	{
		// final LogFileWriter logWriter = new LogFileWriter();
		final LogFileWriter logWriter = new LogFileWriter("syn_iterN.csv", false, true);

		RealWorldDataBenchTestLP.writeHeadline(logWriter);

		final LogFileWriter logWriterTotal = new LogFileWriter("syn_iterN_total.csv", false, true);
		// final LogFileWriter logWriterTotal = new LogFileWriter();
		RealWorldDataBenchTestLP.writeHeadline(logWriterTotal);

		final int[] maxDistValues = { 3, 4, 5 };

		for (final int maxDistValue : maxDistValues)
		{
			for (int n = 80; n <= 150; n += 10)
			// final int n = 100;
			{
				final int[][][] ultrametricArr = SyntheticBenchmarks.getUltrametrics(n, maxDistValue);
				int noSuccessSteps = 0;

				for (int k = 1; k <= ((n * (n - 1)) / 2); k += 3)
				{
					final BenchResult totalSearchResultWithLP = new BenchResult();
					final BenchResult totalSearchResultWithoutLP = new BenchResult();
					final BenchResult totalApproxResult = new BenchResult();
					final BenchResult totalLPResult = new BenchResult();
					final BenchResult totalIlpResult = new BenchResult();

					for (int instanceIter = 0; instanceIter < SyntheticBenchmarks.NO_ITERATIONS; instanceIter++)
					{
						final BenchResult totalPerInstanceSearchResultWithLP = new BenchResult();
						final BenchResult totalPerInstanceApproxResult = new BenchResult();
						final BenchResult totalPerInstanceLPResult = new BenchResult();
						final BenchResult totalPerInstanceIlpResult = new BenchResult();
						final BenchResult totalPerInstanceSearchResultWithoutLP = new BenchResult();

						for (int t = 0; t < SyntheticBenchmarks.NO_ITERATIONS; t++)
						{
							// System.out.println("\tn:" + n + "_k:" + k + "_t:"
							// + t);
							final SymmetricDiscreteDistanceFunction dist = new SymmetricDiscreteDistanceFunction(
									RandomDistanceFunctions.derangedUltrametric(
											SymmetricDiscreteDistanceFunction.copyArray(ultrametricArr[instanceIter]), maxDistValue, k));

							final long begin = System.currentTimeMillis();
							// System.out.println("\truntests");
							final BenchResult[] results = RealWorldDataBenchTestLP.runTests(dist, true, true, true);

							if (((System.currentTimeMillis() - begin) / 1000d) > (RealWorldDataBenchTestLP.maxNumberSeconds / 2))
							{
								noSuccessSteps++;
								System.out.println("\tincrease:" + noSuccessSteps);
							}

							totalPerInstanceApproxResult.add(results[0]);
							totalPerInstanceLPResult.add(results[1]);
							totalPerInstanceSearchResultWithLP.add(results[2]);
							totalPerInstanceIlpResult.add(results[3]);
							totalPerInstanceSearchResultWithoutLP.add(results[4]);
						}
						totalApproxResult.add(totalPerInstanceApproxResult);
						totalLPResult.add(totalPerInstanceLPResult);
						totalSearchResultWithLP.add(totalPerInstanceSearchResultWithLP);
						totalIlpResult.add(totalPerInstanceIlpResult);
						totalSearchResultWithoutLP.add(totalPerInstanceSearchResultWithoutLP);

						RealWorldDataBenchTestLP.writeResults(totalPerInstanceApproxResult, totalPerInstanceLPResult,
								totalPerInstanceSearchResultWithLP, totalPerInstanceIlpResult, totalPerInstanceSearchResultWithoutLP,
								String.valueOf(n) + "\t" + String.valueOf(k) + "\t" + maxDistValue, logWriter);
					}
					RealWorldDataBenchTestLP.writeResults(totalApproxResult, totalLPResult, totalSearchResultWithLP, totalIlpResult,
							totalSearchResultWithoutLP, String.valueOf(n) + "\t" + String.valueOf(k) + "\t" + maxDistValue, logWriterTotal);

					// if ( noSuccessSteps > 10 && k > (int) (1.2 * n) )
					if (noSuccessSteps > 10)
					{
						k = Integer.MAX_VALUE;
						logWriter.flush();
						logWriterTotal.flush();
						break;
					}

					logWriter.flush();
					logWriterTotal.flush();
				}
			}
		}
		logWriter.close();
		logWriterTotal.close();
	}

	public static String toString(final double[] a)
	{
		if (a == null)
			return "null";
		final int iMax = a.length - 1;
		if (iMax == -1)
			return "[]";

		final StringBuilder b = new StringBuilder();
		b.append('[');
		for (int i = 0;; i++)
		{
			if (a[i] != TimeBoundedApplication.STOP_TIME_VALUE)
			{
				b.append(SyntheticBenchmarks.timeFormatter.format(a[i]));
			} else
			{
				b.append("infin");
			}
			if (i == iMax)
				return b.append(']').toString();
			b.append(", ");
		}
	}
}
