package dataReductionRules;

import java.util.BitSet;
import java.util.Observable;

import dataReductionRules.rule3HelpClasses.Clique;
import dataReductionRules.rule3HelpClasses.CliqueGraph;
import dataReductionRules.rule3HelpClasses.SparseIntegerSet;
import dataReductionRules.rule3HelpClasses.ThresholdGraph;
import distanceFunction.SymmetricDiscreteDistanceFunction;
import distanceFunction.SymmetricDiscreteDistanceFunction.DistanceFunctionChange;

/**
 * 
 * @author Tomasz Przedmojski
 * 
 */
public class Rule3Basic extends ObservingAdvancedRule
{

	private SparseIntegerSet allNodesSet;

	private CliqueGraph[] cliqueGraph;
	private ChangeLogger knownInstance = null;
	private int maxValue;
	private ThresholdGraph[] thresholdGraph;

	public Rule3Basic(final int interleavingConst)
	{
		super(interleavingConst);
	}

	private boolean absorbNeighbors(final ChangeLogger instance, final Clique clique, final int t)
	{
		final int maxSize = this.allNodesSet.size();
		boolean changed = false;

		// System.out.println("Absorbing in t="+t+" neighbors of "+clique.getNumber()+" (size="+clique.size()+",|N|="+cliqueGraph[t].neighborCount(clique)+")");

		// compute closedFirstNeighborhood
		final SparseIntegerSet closedFirstNeighborhood = new SparseIntegerSet(maxSize);

		for (final Integer node : clique)
		{
			assert node < maxSize;
			closedFirstNeighborhood.add(node);
		}

		for (final Integer neighbor : this.thresholdGraph[t].neighborsOf(clique.any()))
		{
			closedFirstNeighborhood.add(neighbor);
		}

		final SparseIntegerSet secondNeighborhood = new SparseIntegerSet(maxSize);
		for (final Integer neighbor : this.thresholdGraph[t].neighborsOf(clique.any()))
		{
			for (final Integer secondNeighbor : this.thresholdGraph[t].neighborsOf(neighbor))
			{
				if (!closedFirstNeighborhood.contains(secondNeighbor))
				{
					secondNeighborhood.add(secondNeighbor);
				}
			}

		}

		// assert(secondNeighborhood.size() < clique.size());
		// assert(closedFirstNeighborhood.size() > clique.size());

		final int[] closedFirstNeighborhoodArray = closedFirstNeighborhood.getDenseArray();
		final int[] secondNeighborhoodArray = secondNeighborhood.getDenseArray();

		// do the absorbing
		final SymmetricDiscreteDistanceFunction function = instance.getDistanceFunction();
		int currentDistance;

		// iterate over all pairs
		for (int i = 0; i < closedFirstNeighborhood.size(); ++i)
		{
			for (int j = i + 1; j < closedFirstNeighborhood.size(); ++j)
			{

				// assert(closedFirstNeighborhoodArray[i] < maxSize);
				// assert(closedFirstNeighborhoodArray[j] < maxSize);

				// change the distance only if the current distance > t
				currentDistance = function.get(closedFirstNeighborhoodArray[i], closedFirstNeighborhoodArray[j]);
				if (currentDistance > t)
				{
					changed = true;
					instance.changeDistanceValue(closedFirstNeighborhoodArray[i], closedFirstNeighborhoodArray[j], t - currentDistance, true);
				}
			}
		}

		// iterate over all combinations first : second
		for (int i = 0; i < closedFirstNeighborhood.size(); ++i)
		{
			for (int j = 0; j < secondNeighborhood.size(); ++j)
			{

				// assert(closedFirstNeighborhoodArray[i] <maxSize);
				// assert(secondNeighborhoodArray[j] < maxSize);

				currentDistance = function.get(closedFirstNeighborhoodArray[i], secondNeighborhoodArray[j]);
				// change the distance only if the current distance <= t
				if (currentDistance <= t)
				{
					changed = true;
					instance.changeDistanceValue(closedFirstNeighborhoodArray[i], secondNeighborhoodArray[j], (t + 1) - currentDistance, true);
				}
			}
		}
		// FIXME: isn't it always true?

		// assert(cliqueGraph[t].neighborCount(cliqueGraph[t].getCliqueOfNodeInThresholdGraph(closedFirstNeighborhoodArray[0]))
		// == 0);

		return changed;
	}

	private int calculateMaximumDistanceFromNeighbors(final Clique clique, final int t)
	{
		int maxDist = 0;
		int curDist = 0;
		int preciseDist = 0;
		for (final int node : clique)
		{
			for (final int neighbor : this.thresholdGraph[t].neighborsOf(node))
			{
				if (!clique.equals(neighbor) && (node != neighbor))
				{

					// assert knownInstance != null;
					// assert neighbor <
					// knownInstance.getDistanceFunction().getNoOfElements()
					// : neighbor + " >= " +
					// knownInstance.getDistanceFunction().getNoOfElements();
					// assert node <
					// knownInstance.getDistanceFunction().getNoOfElements()
					// : node + " >= " +
					// knownInstance.getDistanceFunction().getNoOfElements();
					//
					curDist = this.knownInstance.getDistanceFunction().get(node, neighbor);
					if (curDist > maxDist)
					{
						maxDist = curDist;
					}
					preciseDist += curDist;
				}
			}
		}
		// return maxDist;
		return (int) Math.ceil(preciseDist / clique.size());
	}

	private boolean cliqueRemoved(final ChangeLogger instance, final Clique clique, final int t)
	{
		return instance.getDistanceFunction().isDeletedData(clique.any());
	}

	private boolean criticalClique(final SparseIntegerSet nodes, final ChangeLogger instance, final int t)
	{
		Clique clique;
		boolean changed = false;
		while ((clique = this.findPossibleCriticalClique(nodes, instance, t)) != null)
		{
			changed = changed | this.absorbNeighbors(instance, clique, t);
		}
		return changed;
	}

	private Clique findPossibleCriticalClique(final SparseIntegerSet nodes, final ChangeLogger instance, final int t)
	{
		for (final Clique clique : this.cliqueGraph[t])
		{
			// we are only interested in cliques whose elements intersect
			// +nodes+ set
			if (nodes.contains(clique.any()))
			{
				// TODO: order is important for speed!!!
				if (this.firstConditionApply(clique, t))
				{
					// // TRICK: Lemma disallows breaking this CC
					for (final int u : clique)
					{
						for (final int v : clique)
						{
							if (u < v)
							{
								if (instance.getDistanceFunction().get(u, v) == t)
								{
									instance.getModStatistic().avoidIncrease(u, v);
								}
							}
						}
					}
					if (this.secondConditionApply(clique, t))
					{
						// FIXME: remove the if condition?
						if (!this.cliqueRemoved(instance, clique, t))
							return clique;
					}
				}
			}
		}
		return null;
	}

	private boolean firstConditionApply(final Clique clique, final int t)
	{
		final int neighborsCount = this.firstNeighborhoodElementCount(clique, t);

		return (neighborsCount != 0) && (clique.size() >= (t * neighborsCount));
	}

	private int firstNeighborhoodElementCount(final Clique clique, final int t)
	{
		return this.thresholdGraph[t].neighborsOf(clique.any()).size() - clique.size();
	}

	private void handleChange(final DistanceFunctionChange change)
	{
		// System.out.println(change.toString());

		this.updateThresholdGraphs(change);
		this.updateCliqueGraphs(change);
	}

	@Override
	public boolean performReductionRule(final ChangeLogger instance) throws UnsolvableException
	{
		// System.out.println("Applying rule 3");

		this.regenerateGraphsIfNecessary(instance);

		final boolean changed = this.performRule(this.allNodesSet, instance, instance.getDistanceFunction().getMaxValue() - 1);

		if (changed)
		{
			// System.out.println("Rule 3 was useful");
		} else
		{
			// System.out.println("Done nothing");
		}
		return changed;
	}

	private boolean performRule(final SparseIntegerSet nodes, final ChangeLogger instance, final int t) throws UnsolvableException
	{
		if (t <= 0)
			return false;

		boolean changed = this.criticalClique(nodes, instance, t);

		for (final Clique clique : this.cliqueGraph[t])
		{
			// we are only interested in cliques whose elements intersect
			// +nodes+ set
			if (nodes.contains(clique.any()))
			{
				// futhermore they have to be isolated.
				// and not removed
				if ((this.cliqueGraph[t].neighborCount(clique) == 0) && !this.cliqueRemoved(instance, clique, t))
				{
					// is this clique an ultrametric?
					if (!instance.getDistanceFunction().isUltrametric(clique.toVector()))
					{
						changed = changed | this.performRule(clique, instance, t - 1);
					} else
					{
						// yes - remove it
						this.removeCliqueFromInstance(instance, clique);
						changed = true;
					}

				}
			}
		}
		return changed;
	}

	private void regenerateCliqueGraphs()
	{
		this.cliqueGraph = new CliqueGraph[this.maxValue + 1];
		for (int i = 0; i <= this.maxValue; ++i)
		{
			this.cliqueGraph[i] = new CliqueGraph(this.thresholdGraph[i]);
		}
	}

	private void regenerateGraphsIfNecessary(final ChangeLogger instance)
	{
		if (this.knownInstance != instance)
		{
			this.knownInstance = instance;
			this.maxValue = instance.getDistanceFunction().getMaxValue();

			this.regenerateThresholdGraphs();
			this.regenerateCliqueGraphs();
			this.regenerateNodesSet();
		}

	}

	private void regenerateNodesSet()
	{
		this.allNodesSet = new SparseIntegerSet(this.knownInstance.getDistanceFunction().getNoOfElements());

		for (int i = 0; i < this.knownInstance.getDistanceFunction().getNoOfElements(); ++i)
		{
			this.allNodesSet.add(i);
		}
	}

	private void regenerateThresholdGraphs()
	{
		this.thresholdGraph = new ThresholdGraph[this.maxValue + 1];
		for (int i = 0; i <= this.maxValue; ++i)
		{
			this.thresholdGraph[i] = new ThresholdGraph(this.knownInstance, i);
		}

	}

	private void removeCliqueFromInstance(final ChangeLogger instance, final Clique clique)
	{
		for (final Integer node : clique)
		{
			instance.deleteData(node);
		}
	}

	private boolean secondConditionApply(final Clique clique, final int t)
	{
		final int neighborsCount = this.firstNeighborhoodElementCount(clique, t);
		int secondNeighborsCount = 0;

		final BitSet seen = new BitSet(this.maxValue);
		final SymmetricDiscreteDistanceFunction distFunc = this.knownInstance.getDistanceFunction();

		// int maxDist = t;
		// manual calculation of maxDist
		int maxDist = 0;
		int curDist = 0;

		int dist = 0;
		int distCount = 0;

		for (final Clique neighbor : this.cliqueGraph[t].neighborsOf(clique))
		{
			for (final Clique secondNeighbor : this.cliqueGraph[t].neighborsOf(neighbor))
			{
				for (final int secondNeighborNode : secondNeighbor)
				{
					if (!seen.get(secondNeighborNode) && !clique.contains(secondNeighborNode)
							&& !this.thresholdGraph[t].containsEdge(clique.any(), secondNeighborNode))
					{

						seen.set(secondNeighborNode);
						for (final int neighborNode : neighbor)
						{

							curDist = distFunc.get(neighborNode, secondNeighborNode);
							if (curDist > maxDist)
							{
								maxDist = curDist;
							}
							dist += curDist;
						}
					}
				}
			}
			distCount += neighbor.size();
		}

		secondNeighborsCount = seen.cardinality();

		dist = (int) Math.ceil(dist / distCount);

		final boolean ret = clique.size() >= (neighborsCount + (t * secondNeighborsCount));
		return ret;
	}

	@Override
	public void update(final Observable o, final Object data)
	{
		if (this.knownInstance != null)
		{
			/* FIXME: this code is unsafe */
			final SymmetricDiscreteDistanceFunction instance = (SymmetricDiscreteDistanceFunction) o;
			assert instance == this.knownInstance.getDistanceFunction();

			/* FIXME: this code is also unsafe */
			final DistanceFunctionChange change = (DistanceFunctionChange) data;

			this.handleChange(change);
		}
	}

	private void updateCliqueGraphs(final DistanceFunctionChange change)
	{
		// negative change means lowering threshold -> thus adding edges in
		// "lower" graphs
		if (change.isNegative())
		{
			for (int t = change.a; t < change.b; ++t)
			{
				assert t <= this.maxValue;
				assert t >= 0;
				this.cliqueGraph[t].edgeInThresholdGraphAdded(change.i, change.j);
			}
		} else if (change.isPositive())
		{
			for (int t = change.b; t < change.a; ++t)
			{
				assert t <= this.maxValue;
				assert t >= 0;
				this.cliqueGraph[t].edgeInThresholdGraphRemoved(change.i, change.j);
			}
		}
	}

	private void updateThresholdGraphs(final DistanceFunctionChange change)
	{
		// negative change means lowering threshold -> thus adding edges in
		// "lower" graphs
		if (change.isNegative())
		{
			for (int t = change.a; t < change.b; ++t)
			{
				assert t <= this.maxValue;
				assert t >= 0;
				assert t < this.thresholdGraph.length;
				this.thresholdGraph[t].addEdge(change.i, change.j);
			}
		} else if (change.isPositive())
		{
			for (int t = change.b; t < change.a; ++t)
			{
				assert t <= this.maxValue;
				assert t >= 0;
				assert t < this.thresholdGraph.length;
				this.thresholdGraph[t].removeEdge(change.i, change.j);
			}
		}
	}

}
