package dataReductionRules;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Observable;

import distanceFunction.SymmetricDiscreteDistanceFunction;

public class Rule7 extends ObservingAdvancedRule
{

	public Rule7(final int interleavingConst)
	{
		super(interleavingConst);
	}

	private Iterable<Integer> getNonNeighbors(final ArrayList<Integer> neigh, final int numberOfElements)
	{
		class T implements Iterable<Integer>
		{

			@Override
			public Iterator<Integer> iterator()
			{
				class Tmp implements Iterator<Integer>
				{
					private final int n;
					private final ArrayList<Integer> neigh;
					int neighIndex;
					int nextIndex;

					public Tmp(final ArrayList<Integer> neigh, final int numberOfElements)
					{
						this.neigh = neigh;
						this.n = numberOfElements;

						this.neighIndex = 0;
						this.nextIndex = 0;
						this.ensureConsistency();
					}

					public void ensureConsistency()
					{
						while (this.nextIndex == this.neigh.get(this.neighIndex))
						{
							this.nextIndex++;
							if (this.neighIndex < (this.neigh.size() - 1))
							{
								this.neighIndex++;
							}
						}
					}

					@Override
					public boolean hasNext()
					{
						return this.nextIndex < this.n;
					}

					@Override
					public Integer next()
					{
						final int index = this.nextIndex;
						this.nextIndex++;
						this.ensureConsistency();
						return index;
					}

					@Override
					public void remove()
					{
						throw new IllegalAccessError();
					}
				}
				return new Tmp(neigh, numberOfElements);
			}
		}
		return new T();
	}

	private ArrayList<Integer> isReducable(final SymmetricDiscreteDistanceFunction distanceFunction, final int indexI, final int M)
	{
		final ArrayList<Integer> neigh = new ArrayList<>();
		{
			for (int j = 0; j < distanceFunction.getNoOfElements(); j++)
			{
				if ((indexI == j) || (distanceFunction.get(indexI, j) <= M))
				{
					neigh.add(j);
				}
			}
		}

		int delta_V = 0;
		{
			for (int i = 0; i < neigh.size(); i++)
			{
				for (int j = i + 1; j < neigh.size(); j++)
				{
					if (distanceFunction.get(neigh.get(i), neigh.get(j)) > M)
					{
						delta_V++;
					}
				}
			}
		}

		if ((2 * delta_V) >= neigh.size())
			return null;

		int gamma_v = 0;
		{
			for (final int indexU : neigh)
			{
				for (final int indexW : this.getNonNeighbors(neigh, distanceFunction.getNoOfElements()))// iterate over all non-neigh
				{
					gamma_v += (M + 1) - distanceFunction.get(indexU, indexW);
				}
				if (((2 * delta_V) + gamma_v) >= neigh.size())
					return null;
			}
		}
		return neigh;
	}

	@Override
	public boolean performReductionRule(final ChangeLogger solver) throws UnsolvableException
	{
		// long time = System.currentTimeMillis();
		boolean somethingChanged = false;
		final SymmetricDiscreteDistanceFunction distanceFunction = solver.getDistanceFunction();
		final int M = distanceFunction.getMaxValue() - 1;
		for (int indexV = 0; indexV < distanceFunction.getNoOfElements(); indexV++)
		{
			ArrayList<Integer> neigh;
			if ((neigh = this.isReducable(distanceFunction, indexV, M)) != null)
			{
				{// decrease M+1 values in N_v to M
					for (int indexI = 0; indexI < neigh.size(); indexI++)
					{
						for (int indexJ = indexI + 1; indexJ < neigh.size(); indexJ++)
						{
							final int i = neigh.get(indexI);
							final int j = neigh.get(indexJ);
							final int value = solver.getDistanceFunction().get(i, j);
							if (value > M)
							{
								solver.changeDistanceValue(i, j, M - value, true);
								// System.err.println("decrease d(" + i + "," + j + ") by " + (M - value) + " ("
								// + solver.getDistanceFunction().get(i, j) + ")");
								somethingChanged = true;
							}
						}
					}
				}
				{// increase distance to vertices outside of N_v
					for (final int nonNeighX : this.getNonNeighbors(neigh, solver.getDistanceFunction().getNoOfElements()))
					{
						int sum = 0;
						for (final int n : neigh)
						{
							sum += (M + 1) - solver.getDistanceFunction().get(nonNeighX, n);
						}
						if (sum <= (int) Math.ceil(neigh.size() / 2d))
						{
							for (final int n : neigh)
							{
								final int value = solver.getDistanceFunction().get(nonNeighX, n);
								if (value < (M + 1))
								{
									solver.changeDistanceValue(n, nonNeighX, (M + 1) - value, true);
									// System.err.println("increase d(" + nonNeighX + "," + n + ") by " + ((M + 1) - value) + " ("
									// + solver.getDistanceFunction().get(nonNeighX, n) + ")");
									somethingChanged = true;
								}
							}
						}
					}
				}
			}
		}
		// if (somethingChanged)
		// System.out.println("rule 7(" + somethingChanged + "," + (System.currentTimeMillis() - time) / 1000d + "s)");
		return somethingChanged;
	}

	@Override
	public void update(final Observable o, final Object arg)
	{
		// TODO Auto-generated method stub

	}
}
