package dataReductionRules.rule3HelpClasses;

import java.util.Iterator;

/**
 * 
 * @author Tomasz Przedmojski
 */
public class SparseIntegerIdenfitiableSet<Node extends IntegerIdentifiable> implements Iterable<Node>
{
	protected Node dense[];
	protected int members;
	protected int sparse[];

	/**
	 * Constructs the set. Performance guarantee: O(n)
	 * 
	 * @param maxValue (n) maximum possible value in set.
	 */
	public SparseIntegerIdenfitiableSet(final int maxValue)
	{
		this.sparse = new int[maxValue + 1];
		this.dense = (Node[]) new IntegerIdentifiable[maxValue + 1];
		for (int i = 0; i <= maxValue; ++i)
		{
			this.dense[i] = null;
		}
		this.members = 0;
	}

	/**
	 * Adds element with given value to the set. Performance guarantee: O(1)
	 * 
	 * @param value value of the element
	 * @return true if this set did not already contain the specified element.
	 */
	public boolean add(final Node value)
	{
		if (!this.contains(value))
		{
			this.sparse[value.getNumber()] = this.members;
			this.dense[this.members] = value;
			this.members++;

			return true;
		}
		return false;
	}

	/**
	 * Removes all elements from the set. Performance guarantee: O(1)
	 */
	public void clear()
	{
		this.members = 0;
	}

	/**
	 * Checks if element with given value is present in set. Performance guarantee: O(1)
	 * 
	 * @param value value of the element
	 * @return true if element is present, false otherwise.
	 */
	public boolean contains(final Node value)
	{

		if (value.getNumber() >= this.sparse.length)
			throw new RuntimeException("clique=" + value.getNumber() + " but max index=" + this.sparse.length);

		final int k = this.sparse[value.getNumber()];
		return (k < this.members) && this.dense[k].equals(value);
	}

	public Object[] getDenseArray()
	{
		return this.dense;
	}

	/**
	 * Checks if any elements are present in set. Performance guarantee: O(1)
	 * 
	 * @return true if set is not empty, false otherwise.
	 */
	public boolean isEmpty()
	{
		return this.members == 0;
	}

	@Override
	public Iterator<Node> iterator()
	{
		return new Iterator<Node>()
		{
			int index = 0;

			@Override
			public boolean hasNext()
			{
				return this.index < SparseIntegerIdenfitiableSet.this.members;
			}

			@Override
			public Node next()
			{
				return SparseIntegerIdenfitiableSet.this.dense[this.index++];
			}

			@Override
			public void remove()
			{
				throw new UnsupportedOperationException("Not supported.");
			}
		};
	}

	/**
	 * Removes element with given value from the set. For elements not present in the set this method is idempotent. Performance guarantee: O(1)
	 * 
	 * @param value value of the element
	 * @return true if this set did contain the specified element.
	 */
	public boolean remove(final Node value)
	{
		if (!this.contains(value))
			return false;
		else
		{
		}

		if ((this.members == 1) || (this.sparse[value.getNumber()] == this.members))
		{
			this.members--;
		} else
		{
			final int freedIndex = this.sparse[value.getNumber()];
			final Node newValue = this.dense[this.members - 1];

			this.sparse[newValue.getNumber()] = freedIndex;
			this.dense[freedIndex] = newValue;
			this.members--;
		}
		return true;
	}

	/**
	 * Returns the number of elements in the set. Use in conjuction with getDenseArray() Performance guarantee: O(1)
	 */
	public int size()
	{
		return this.members;
	}

}
