package solver.heuristics;

import java.util.Map;
import java.util.TreeMap;

import solver.hierarchicalClustering.HierarchicalClusteringSolver;
import distanceFunction.SymmetricDiscreteDistanceFunction;
import distanceFunction.random.MergeElements;

/**
 * This class provides an implementation of the (single,complete,average) linkage algorithm.
 * 
 * @author Sepp Hartung
 * 
 */
public class LinkageSolver extends HierarchicalClusteringSolver
{

	public static final int AVERAGE_LINKAGE = 3;

	public static final int COMPLETE_LINKAGE = 2;

	public static final int SINGLE_LINKAGE = 1;

	private final Map<Integer, MergeElements> mergedElements;

	private final int method;

	public LinkageSolver(final SymmetricDiscreteDistanceFunction dist, final int param, final int linkageMethod)
	{
		super(dist, param);
		this.method = linkageMethod;
		this.mergedElements = new TreeMap<Integer, MergeElements>();
	}

	private void changeDistance(final int i, final int j, final int value)
	{
		this.decreaseParameter(this.mergedElements.get(i).setValue(this, this.mergedElements.get(j), value));
	}

	private int[] getMinDistance()
	{
		int dist = Integer.MAX_VALUE;
		final int[] indices = { -1, -1 };

		for (int i = 0; i < this.getNoOfElements(); i++)
		{
			if (!this.isDeletedData(i))
			{
				for (int j = i + 1; j < this.getNoOfElements(); j++)
				{
					if (!this.isDeletedData(j) && (this.get(i, j) < dist))
					{
						indices[0] = i;
						indices[1] = j;
						dist = this.get(i, j);
					}
				}
			}
		}
		return indices;
	}

	private void reComputeDistances(final int[] min)
	{
		for (int i = 0; i < this.getNoOfElements(); i++)
		{
			final SymmetricDiscreteDistanceFunction dist = this;
			if (!dist.isDeletedData(i) && (i != min[0]) && (i != min[1]))
			{
				final int d1 = dist.get(min[0], i);
				final int d2 = dist.get(min[1], i);
				if (this.method == LinkageSolver.SINGLE_LINKAGE)
				{
					final int l = Math.min(d1, d2);
					this.changeDistance(min[0], i, l);
				} else if (this.method == LinkageSolver.COMPLETE_LINKAGE)
				{
					this.changeDistance(min[0], i, Math.max(d1, d2));
				} else if (this.method == LinkageSolver.AVERAGE_LINKAGE)
				{
					this.changeDistance(min[0], i, Math.round((d1 + d2) / 2f));
				}
			}
		}
	}

	@Override
	public boolean solve()
	{
		for (int i = 0; i < this.getNoOfElements(); i++)
		{
			this.mergedElements.put(i, new MergeElements(i));
		}
		for (int i = 1; i < this.getNoOfElements(); i++)
		{
			final int[] min = this.getMinDistance();

			// delete node
			this.deleteData(min[1]);
			this.mergedElements.put(min[0], new MergeElements(this.mergedElements.get(min[0]), this.mergedElements.remove(min[1])));

			this.reComputeDistances(min);
		}
		return false;
	}
}