package solver.hierarchicalClustering;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction;
import distanceFunction.SymmetricDiscreteDistanceFunction;

/**
 * Store and gives comfortable access to all conflicts of a distance function. This class will be used to (ini) compute all conflicts for a given
 * distance function.
 * 
 * @author Sepp Hartung
 * @version 1.0
 * 
 */
public class Conflicts
{
	/**
	 * Checks whether the conflict contains the edge i,j. If so, it returns the third element; otherwise -1;
	 * 
	 * @param conflict
	 * @param i
	 * @param j
	 * @return
	 */
	public static int containsEdge(final int[] conflict, final int i, final int j)
	{
		if (conflict[0] == i)
		{
			if (conflict[1] == j)
				return conflict[2];
			else if (conflict[2] == j)
				return conflict[1];
		} else if (conflict[1] == i)
		{
			if (conflict[0] == j)
				return conflict[2];
			else if (conflict[2] == j)
				return conflict[0];
		} else if (conflict[2] == i)
		{
			if (conflict[0] == j)
				return conflict[1];
			else if (conflict[1] == j)
				return conflict[0];
		}
		return -1;
	}

	private final SymmetricDiscreteDistanceFunction conflictEdge;

	/**
	 * Store of conflicts.
	 */
	private final HashSet<ConflictArr> conflicts;

	private final SymmetricDiscreteDistanceFunction nonConflictEdge;

	private Conflicts(final HashSet<ConflictArr> conflicts, final SymmetricDiscreteDistanceFunction conflictEdge,
			final SymmetricDiscreteDistanceFunction nonConflictEdge)
	{
		this.conflicts = conflicts;
		this.conflictEdge = conflictEdge;
		this.nonConflictEdge = nonConflictEdge;
	}

	public Conflicts(final HierarchicalClusteringSolver solver)
	{
		// this.solver = solver;
		this.conflicts = new HashSet<ConflictArr>();
		this.conflictEdge = new SymmetricDiscreteDistanceFunction(solver.getNoOfElements());
		this.nonConflictEdge = new SymmetricDiscreteDistanceFunction(solver.getNoOfElements());

		{// ini conflicts O(n^3)
			for (int i = 0; i < solver.getNoOfElements(); i++)
			{
				if (!solver.isDeletedData(i))
				{
					for (int j = i + 1; j < solver.getNoOfElements(); j++)
					{
						if (!solver.isDeletedData(j))
						{
							for (int l = j + 1; l < solver.getNoOfElements(); l++)
							{
								if (!solver.isDeletedData(l))
								{ // CONFLICT?
									int[] indiceArr;
									if ((indiceArr = solver.isConflict(i, j, l)) != null)
									{
										this.conflicts.add(new ConflictArr(indiceArr));

										final int minEffort = solver.get(indiceArr[0], indiceArr[1])
												- Math.max(solver.get(indiceArr[0], indiceArr[2]), solver.get(indiceArr[1], indiceArr[2]));
										this.conflictEdge.increase(indiceArr[0], indiceArr[1], minEffort);
										this.nonConflictEdge.increase(indiceArr[0], indiceArr[2], minEffort);
										this.nonConflictEdge.increase(indiceArr[1], indiceArr[2], minEffort);
									}
								}
							}
						}
					}
				}
			}
		}
	}

	@Override
	public Conflicts clone()
	{
		return new Conflicts(new HashSet<ConflictArr>(this.conflicts), new SymmetricDiscreteDistanceFunction(this.conflictEdge),
				new SymmetricDiscreteDistanceFunction(this.nonConflictEdge));
	}

	public ReadOnlyViewOnSymmetricDiscreteDistanceFunction getConflictEdgeStatistic()
	{
		return this.conflictEdge.getUnmodifiableView();
	}

	public Collection<int[]> getConflicts()
	{
		class TmpView implements Collection<int[]>
		{
			@Override
			public boolean add(final int[] e)
			{
				throw new IllegalAccessError();
			}

			@Override
			public boolean addAll(final Collection<? extends int[]> c)
			{
				throw new IllegalAccessError();
			}

			@Override
			public void clear()
			{
				throw new IllegalAccessError();
			}

			@Override
			public boolean contains(final Object o)
			{
				throw new IllegalAccessError();
			}

			@Override
			public boolean containsAll(final Collection<?> c)
			{
				throw new IllegalAccessError();
			}

			@Override
			public boolean isEmpty()
			{
				throw new IllegalAccessError();
			}

			@Override
			public Iterator<int[]> iterator()
			{
				class IterConf implements Iterator<int[]>
				{
					private final Iterator<ConflictArr> iter;

					public IterConf()
					{
						this.iter = Conflicts.this.conflicts.iterator();
					}

					@Override
					public boolean hasNext()
					{
						return this.iter.hasNext();
					}

					@Override
					public int[] next()
					{
						return this.iter.next().con;
					}

					@Override
					public void remove()
					{
						throw new IllegalAccessError();
					}
				}
				return new IterConf();
			}

			@Override
			public boolean remove(final Object o)
			{
				throw new IllegalAccessError();
			}

			@Override
			public boolean removeAll(final Collection<?> c)
			{
				throw new IllegalAccessError();
			}

			@Override
			public boolean retainAll(final Collection<?> c)
			{
				throw new IllegalAccessError();
			}

			@Override
			public int size()
			{
				return Conflicts.this.conflicts.size();
			}

			@Override
			public Object[] toArray()
			{
				throw new IllegalAccessError();
			}

			@Override
			public <T> T[] toArray(final T[] a)
			{
				throw new IllegalAccessError();
			}

		}
		return new TmpView();
	}

	public ReadOnlyViewOnSymmetricDiscreteDistanceFunction getNonConflictEdgeStatistic()
	{
		return this.nonConflictEdge.getUnmodifiableView();
	}

	public boolean isEmpty()
	{
		return this.conflicts.isEmpty();
	}

	@Override
	public String toString()
	{
		final StringBuffer s = new StringBuffer();
		s.append('[');
		final Iterator<ConflictArr> iter = this.conflicts.iterator();
		while (iter.hasNext())
		{
			s.append(iter.next());
		}
		s.append(']');
		return s.toString();
	}

	/**
	 * Computes all conflicts which contains the edge between i and j.
	 * 
	 * @param i
	 * @param j
	 * @param c
	 * @return
	 */
	public void updateConflictsForEdge(final SymmetricDiscreteDistanceFunction distanceFunction, final int i, final int j, final int c)
	{
		// final Iterator<ConflictArr> iter = conflicts.iterator();
		// while (iter.hasNext())
		// {
		// final int[] conflict = iter.next().con;
		// final int l;
		// if ((l = Conflicts.containsEdge(conflict, i, j)) >= 0)
		// {
		// iter.remove();
		//
		// final int[] distanceValues = { distanceFunction.get(i, j) - c,
		// distanceFunction.get(i, l), distanceFunction.get(j, l) };
		// Arrays.sort(distanceValues);
		// final int minEffort = distanceValues[1] - distanceValues[2];
		//
		// conflictEdge.increase(conflict[0], conflict[1], minEffort);
		// nonConflictEdge.increase(conflict[1], conflict[2], minEffort);
		// nonConflictEdge.increase(conflict[0], conflict[2], minEffort);
		// }
		// }

		for (int l = 0; l < distanceFunction.getNoOfElements(); l++)
		{
			if ((i != l) && (j != l) && !distanceFunction.isDeletedData(l))
			{
				{// remove old conflict
					final int d_i_j = distanceFunction.get(i, j) - c;
					final int d_i_l = distanceFunction.get(i, l);
					final int d_j_l = distanceFunction.get(j, l);
					final int[] oldConflict = SymmetricDiscreteDistanceFunction.isConflict(i, j, l, d_i_j, d_i_l, d_j_l);

					if (oldConflict != null)
					{
						this.conflicts.remove(new ConflictArr(oldConflict));
						final int[] distanceValues = { d_i_j, d_i_l, d_j_l };
						Arrays.sort(distanceValues);
						final int minEffort = distanceValues[1] - distanceValues[2];

						this.conflictEdge.increase(oldConflict[0], oldConflict[1], minEffort);
						this.nonConflictEdge.increase(oldConflict[1], oldConflict[2], minEffort);
						this.nonConflictEdge.increase(oldConflict[0], oldConflict[2], minEffort);
					}
				}
				// add new conflict
				final int[] conflict = distanceFunction.isConflict(i, j, l);
				if (conflict != null)
				{
					this.conflicts.add(new ConflictArr(conflict));

					final int i1 = conflict[0];
					final int j1 = conflict[1];
					final int l1 = conflict[2];
					final int minEffort = distanceFunction.get(i1, j1) - Math.max(distanceFunction.get(i1, l1), distanceFunction.get(j1, l1));
					this.conflictEdge.increase(i1, j1, minEffort);
					this.nonConflictEdge.increase(j1, l1, minEffort);
					this.nonConflictEdge.increase(i1, l1, minEffort);
				}
			}
		}
	}
}