package solver.hierarchicalClustering;

import java.util.Collection;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Vector;

import solver.Solver;
import dataReductionRules.ChangeLogger;
import dataReductionRules.RuleInterface;
import dataReductionRules.UnsolvableException;
import distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction;
import distanceFunction.SymmetricDiscreteDistanceFunction;

/**
 * Basic class for all solvers.*
 * 
 * @author Sepp Hartung
 * @version 1.0
 * 
 */
public abstract class HierarchicalClusteringSolver extends Solver implements SolverInterface
{
	private final Conflicts conflictStorage;

	public ModificationStatistic modStatistic;

	/**
	 * Constructs a Hierarchical-Clustering solver.
	 * 
	 * @param distanceFunctions distance function (will be copied)
	 * @param conflicts will be computed from scratch
	 */
	public HierarchicalClusteringSolver(final SymmetricDiscreteDistanceFunction dist, final int parameter)
	{
		this(dist, parameter, null);
	}

	public HierarchicalClusteringSolver(final SymmetricDiscreteDistanceFunction dist, final int parameter, final Conflicts conflicts)
	{
		super(dist, parameter);
		this.modStatistic = new ModificationStatistic(this);
		this.conflictStorage = conflicts == null ? new Conflicts(this) : conflicts;
	}

	/**
	 * d_i_j=+c; update conflicts; decrease/increase parameter value
	 * 
	 * @param i
	 * @param j
	 * @param c
	 * @param restoreValue
	 * @throws UnsolvableException
	 */
	public void changeDistanceValue(final int i, final int j, final int c, final boolean decreaseParam)
	{
		if (decreaseParam)
		{
			if (c < 0)
			{
				this.modStatistic.avoidIncrease(i, j);
			} else
			{
				this.modStatistic.avoidDecrease(i, j);
			}
		}
		this.increase(i, j, c);
		if (decreaseParam)
		{
			this.decreaseParameter(Math.abs(c));
		} else
		{
			this.increaseParameter(Math.abs(c));
		}

		// update conflicts
		this.getConflictStorage().updateConflictsForEdge(this, i, j, c);

		if (!decreaseParam)
		{
			if (c > 0)
			{
				this.modStatistic.removeIncreaseBlock(i, j);
			} else
			{
				this.modStatistic.removeDecreaseBlock(i, j);
			}
		}
	}

	/**
	 * Computes a decomposition in independent instances.
	 * 
	 * 1 *
	 * 
	 * @return
	 */
	public SymmetricDiscreteDistanceFunction[] computeDecomposition()
	{
		class ConflictSet
		{
			class Edge implements Comparable<Edge>
			{
				private int i;

				private int j;

				public Edge(final int i, final int j)
				{
					if (i <= j)
					{
						this.i = i;
						this.j = j;
					} else
					{
						this.i = j;
						this.j = i;
					}
				}

				@Override
				public int compareTo(final ConflictSet.Edge edge)
				{
					final int dif = this.i - edge.i;
					if (dif != 0)
						return dif;
					return this.j - edge.j;
				}

				@Override
				public boolean equals(final Object obj)
				{
					if (this == obj)
						return true;
					if (obj == null)
						return false;
					if (this.getClass() != obj.getClass())
						return false;
					final Edge other = (Edge) obj;
					if (this.i != other.i)
						return false;
					if (this.j != other.j)
						return false;
					return true;
				}
			}

			final SortedSet<Edge> edges;;

			public ConflictSet()
			{
				this.edges = new TreeSet<Edge>();
			}

			public void addConflict(final int[] conflict)
			{
				final Edge[] edgesT = this.getDecConflict(conflict);
				for (final Edge edge : edgesT)
				{
					this.edges.add(edge);
				}
			}

			public boolean containsEdge(final int[] conflict)
			{
				final Edge[] edges = this.getDecConflict(conflict);
				for (final Edge edge : edges)
				{
					if (this.edges.contains(edge))
						return true;
				}
				return false;
			}

			public Edge[] getDecConflict(final int[] conflict)
			{
				final Edge arr1 = new Edge(conflict[0], conflict[1]);
				final Edge arr2 = new Edge(conflict[1], conflict[2]);
				final Edge arr3 = new Edge(conflict[0], conflict[2]);
				return new Edge[] { arr1, arr2, arr3 };
			}

			public int[] getNodes()
			{
				final Set<Integer> nodes = new TreeSet<Integer>();
				for (final Edge edge : this.edges)
				{
					nodes.add(edge.i);
					nodes.add(edge.j);
				}
				final int[] out = new int[nodes.size()];
				int i = 0;
				for (final Integer node : nodes)
				{
					out[i++] = node;
				}
				return out;
			}
		}
		;

		final Vector<ConflictSet> isolatedCliques = new Vector<ConflictSet>();
		for (final int[] conflict : this.getConflictStorage().getConflicts())
		{
			boolean flag = false;
			for (final ConflictSet isConflictSet : isolatedCliques)
			{
				if (isConflictSet.containsEdge(conflict))
				{
					isConflictSet.addConflict(conflict);
					flag = true;
				}
			}
			if (!flag)
			{
				final ConflictSet conflictSet = new ConflictSet();
				isolatedCliques.add(conflictSet);
				conflictSet.addConflict(conflict);
			}
		}

		final SymmetricDiscreteDistanceFunction[] distFunctions = new SymmetricDiscreteDistanceFunction[isolatedCliques.size()];
		for (int i = 0; i < isolatedCliques.size(); i++)
		{
			distFunctions[i] = new SymmetricDiscreteDistanceFunction(isolatedCliques.get(i).getNodes(), this);
		}
		return distFunctions;

	}

	@Override
	public ReadOnlyViewOnSymmetricDiscreteDistanceFunction getConflictEdgeStatistic()
	{
		return this.getConflictStorage().getConflictEdgeStatistic();
	}

	@Override
	public Collection<int[]> getConflicts()
	{
		return this.getConflictStorage().getConflicts();
	}

	public Conflicts getConflictStorage()
	{
		return this.conflictStorage;
	}

	@Override
	public SymmetricDiscreteDistanceFunction getDistanceFunction()
	{
		return this;
	}

	@Override
	public ReadOnlyViewOnSymmetricDiscreteDistanceFunction getNonConflictEdgeStatistic()
	{
		return this.getConflictStorage().getNonConflictEdgeStatistic();
	}

	/**
	 * Apply all data reduction rules to the instance.
	 * 
	 * @param reverseChangesIfUnsolvable TODO
	 * @param distanceFunction
	 * 
	 * @return
	 * @throws UnsolvableException
	 */
	public ChangeLogger performDataReductionRules(final boolean reverseChangesIfUnsolvable) throws UnsolvableException
	{
		boolean f = true;

		// apply all rules (param applyRulesExhaustively determines whether they
		// get applied as long as something is removed or if they are only
		// applied once
		final ChangeLogger logger = new ChangeLogger(this);
		try
		{
			while (f && !this.getConflictStorage().isEmpty())
			{
				f = false;
				if (SearchTreeSolver.randomizeRuleOrderEachApplication.getValue())
				{
					this.randomizeRuleOrder();
				}

				for (final RuleInterface rule : this.getRules())
				{
					int numberConflicts = -1;
					if (SearchTreeSolver.RulesVerboseLevel > 0)
					{
						System.out.print("\tstart " + rule.getClass());
						numberConflicts = this.getConflicts().size();
					}
					final boolean reducedSomething = rule.performRule(logger);

					if (SearchTreeSolver.RulesVerboseLevel > 0)
					{
						final int size = this.getConflicts().size();
						System.out.println(" ("
								+ (reducedSomething ? "reduced (#c: " + numberConflicts + " > " + size + " (" + (size - numberConflicts) + ")"
										: "nothing") + ", dist hash code:" + SymmetricDiscreteDistanceFunction.hashSumDist((this)) + ")");
					}

					f = f || reducedSomething;
				}
				if (!SearchTreeSolver.applyRulesExhaustively.getValue())
				{
					f = false;
				}
			}
		} catch (final UnsolvableException e)
		{
			// in case of the instance is not solvable, change back
			// modifications that have been
			// done by the rules
			if (reverseChangesIfUnsolvable)
			{
				logger.reverseAllChanges();
			}
			throw e;
		}
		return logger;
	}
}