/**
 * 
 */
package solver.hierarchicalClustering;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.TreeMap;

import org.apache.commons.lang3.mutable.MutableBoolean;
import org.apache.commons.lang3.mutable.MutableInt;

import solver.lpBased.HierClusterLpSolver;
import dataReductionRules.ChangeLogger;
import dataReductionRules.Rule0;
import dataReductionRules.Rule1;
import dataReductionRules.Rule2;
import dataReductionRules.Rule3;
import dataReductionRules.Rule4;
import dataReductionRules.Rule5;
import dataReductionRules.Rule6;
import dataReductionRules.Rule7;
import dataReductionRules.RuleInterface;
import dataReductionRules.UnsolvableException;
import dataStructures.Store;
import distanceFunction.SymmetricDiscreteDistanceFunction;

/**
 * 3^k search tree implementation.
 * 
 * @author Sepp Hartung
 * @version 1.0
 */
public class SearchTreeSolver extends HierarchicalClusteringSolver
{
	static class ConflictComparator implements Comparator<int[]>
	{
		@Override
		public int compare(final int[] o1, final int[] o2)
		{
			if (o1[0] < o2[0])
				return -1;
			else if (o1[0] > o2[0])
				return 1;
			else if (o1[1] < o2[1])
				return -1;
			else if (o1[1] > o2[1])
				return 1;
			else if (o1[2] < o2[2])
				return -1;
			else if (o1[2] > o2[2])
				return 1;
			return 0;

		}
	}

	public static final int ALMOST_FIXED_SCORE_VALUE = SearchTreeSolver.FIXED_SCORE_VALUE / 2;

	public static final MutableBoolean Apply_Rule0 = new MutableBoolean(true);

	public static final MutableBoolean Apply_Rule1 = new MutableBoolean(true);

	public static final MutableBoolean Apply_Rule2 = new MutableBoolean(true);

	public static final MutableBoolean Apply_Rule3 = new MutableBoolean(true);

	public static final MutableBoolean Apply_Rule4 = new MutableBoolean(true);

	public static final MutableBoolean Apply_Rule5 = new MutableBoolean(true);

	public static final MutableBoolean Apply_Rule6 = new MutableBoolean(true);

	public static final MutableBoolean Apply_Rule7 = new MutableBoolean(true);

	public static final MutableBoolean applyRulesExhaustively = new MutableBoolean(true);

	private static final int FIXED_SCORE_VALUE = Integer.MAX_VALUE;

	public static final MutableInt interleaveDR_Bound_Large = new MutableInt(50);

	public static final MutableInt interleaveDR_Constant_Large = new MutableInt(2);

	public static final MutableInt interleaveDR_Constant_Small = new MutableInt(2);

	public static final MutableInt interleaveLP_Bound_Medium = new MutableInt(50);

	public static int interleaveLP_Bound_Small = 20;

	public static final MutableInt interleaveLP_Constant_Large = new MutableInt(9);

	public static final MutableInt interleaveLP_Constant_Medium = new MutableInt(7);

	public static final MutableInt interleaveLP_Constant_Small = new MutableInt(3);

	public static final MutableInt offset_DR = new MutableInt(0);

	public static final MutableInt offset_LP = new MutableInt(2);

	public static final MutableBoolean randomizeRuleOrderAtBeginning = new MutableBoolean(false);

	public static final MutableBoolean randomizeRuleOrderEachApplication = new MutableBoolean(false);

	public static int RulesVerboseLevel = 0;

	public static int SearchTreeSolver_LogLevel = 0;

	public static final MutableBoolean useLevelForKernelInterleavingConstant = new MutableBoolean(false);

	public static final MutableBoolean useLevelForLPInterleavingConstant = new MutableBoolean(true);

	private static int getScore(final int[] conflict, final ModificationStatistic modStat)
	{
		final int i = conflict[0];
		final int j = conflict[1];
		final int l = conflict[2];
		if (modStat != null)
		{

			if (!modStat.isDecreasePossible(i, j) && (!modStat.isIncreasePossible(i, l) || !modStat.isIncreasePossible(j, l)))
				return SearchTreeSolver.FIXED_SCORE_VALUE;
			if (!modStat.isIncreasePossible(i, l) && !modStat.isIncreasePossible(j, l))
				return SearchTreeSolver.FIXED_SCORE_VALUE;
			if (!modStat.isDecreasePossible(i, j) || !modStat.isIncreasePossible(i, l) || !modStat.isIncreasePossible(j, l))
				return SearchTreeSolver.ALMOST_FIXED_SCORE_VALUE;
		}
		return 1;
	}

	public int almostFixedConflicts;
	public final ConflictComparator conflictComparator = new ConflictComparator();
	public int fixedConflicts;
	private boolean interleaveDR = true;
	private boolean interleaveLP = true;

	/**
	 * Recursive step counter.
	 */
	public long iter = 0;

	private int level;

	private TreeMap<Long, HierClusterLpSolver> lpHistory;

	private int lpSuccess = 0;

	public int normalConflicts;

	public final boolean verbose;

	public static MutableInt Rule0_InterleavingConst = new MutableInt(1);
	public static MutableInt Rule1_InterleavingConst = new MutableInt(1);
	public static MutableInt Rule2_InterleavingConst = new MutableInt(1);
	public static MutableInt Rule3_InterleavingConst = new MutableInt(1);
	public static MutableInt Rule4_InterleavingConst = new MutableInt(1);
	public static MutableInt Rule5_InterleavingConst = new MutableInt(1);
	public static MutableInt Rule6_InterleavingConst = new MutableInt(100);
	public static MutableInt Rule7_InterleavingConst = new MutableInt(100);

	/*
	 * by extremely careful when providing Conflicts (used to avoid multiple recomputation of it, Conflicts have clone function..)
	 */
	public SearchTreeSolver(final Conflicts conflicts, final SymmetricDiscreteDistanceFunction dist,
			final Store<ArrayList<int[]>, ArrayList<int[]>> modBlocks, final int param, final boolean verbose)
	{
		super(dist, param, conflicts);
		{ // add data reduction rules
			if (SearchTreeSolver.Apply_Rule0.getValue())
			{
				this.addRule(new Rule0(Rule0_InterleavingConst.intValue()));
			}
			if (SearchTreeSolver.Apply_Rule1.getValue())
			{
				this.addRule(new Rule1(Rule1_InterleavingConst.intValue()));
			}
			if (SearchTreeSolver.Apply_Rule2.getValue())
			{
				this.addRule(new Rule2(Rule2_InterleavingConst.intValue()));
			}
			if (SearchTreeSolver.Apply_Rule4.getValue())
			{
				this.addRule(new Rule4(Rule4_InterleavingConst.intValue()));
			}
			if (SearchTreeSolver.Apply_Rule5.booleanValue())
			{
				this.addRule(new Rule5(Rule5_InterleavingConst.intValue()));
			}
			//
			if (SearchTreeSolver.Apply_Rule3.getValue())
			{
				this.addObservingRule(new Rule3(Rule3_InterleavingConst.intValue()));
			}
			if (SearchTreeSolver.Apply_Rule6.getValue())
			{
				this.addRule(new Rule6(Rule6_InterleavingConst.intValue()));
			}
			if (SearchTreeSolver.Apply_Rule7.getValue())
			{
				this.addObservingRule(new Rule7(Rule7_InterleavingConst.intValue()));
			}

			if (SearchTreeSolver.randomizeRuleOrderAtBeginning.getValue())
			{
				this.randomizeRuleOrder();
			}
		}
		this.verbose = verbose;

		if ((modBlocks != null) && (modBlocks.getS1() != null))
		{
			for (final int[] increaseBlock : modBlocks.getS1())
			{
				this.modStatistic.avoidIncrease(increaseBlock[0], increaseBlock[1]);
			}
		}
		if ((modBlocks != null) && (modBlocks.getS2() != null))
		{
			for (final int[] decreaseBlock : modBlocks.getS2())
			{
				this.modStatistic.avoidDecrease(decreaseBlock[0], decreaseBlock[1]);
			}
		}
	}

	public SearchTreeSolver(final SymmetricDiscreteDistanceFunction dist, final int param)
	{
		this(dist, param, false);
	}

	public SearchTreeSolver(final SymmetricDiscreteDistanceFunction dist, final int param, final boolean verbose)
	{
		this(dist, null, param, verbose);
	}

	public SearchTreeSolver(final SymmetricDiscreteDistanceFunction dist, final int param, final boolean verbose, final RuleInterface[] rules)
	{
		super(dist, param);
		this.verbose = verbose;
		for (final RuleInterface rule : rules)
		{
			this.addRule(rule);
		}
		if (SearchTreeSolver.randomizeRuleOrderAtBeginning.getValue())
		{
			this.randomizeRuleOrder();
		}
	}

	public SearchTreeSolver(final SymmetricDiscreteDistanceFunction dist, final Store<ArrayList<int[]>, ArrayList<int[]>> modBlocks, final int param)
	{
		this(dist, modBlocks, param, false);
	}

	public SearchTreeSolver(final SymmetricDiscreteDistanceFunction dist, final Store<ArrayList<int[]>, ArrayList<int[]>> modBlocks, final int param,
			final boolean verbose)
	{
		this(null, dist, modBlocks, param, verbose);
	}

	private int[] chooseConflict()
	{
		int maxScore = 0;
		int[] greatestScoreConflict = null;

		for (final int[] conflict : this.getConflictStorage().getConflicts())
		{
			final int conflictScore;
			if (((conflictScore = SearchTreeSolver.getScore(conflict, this.modStatistic)) > maxScore) || (greatestScoreConflict == null)
					|| ((conflictScore == maxScore) && (this.conflictComparator.compare(greatestScoreConflict, conflict) > 0)))
			{
				maxScore = conflictScore;
				greatestScoreConflict = conflict;
				if (maxScore == SearchTreeSolver.FIXED_SCORE_VALUE)
				{
					break;
				}
			}
		}
		if (maxScore == SearchTreeSolver.FIXED_SCORE_VALUE)
		{
			this.fixedConflicts++;
		} else if (maxScore == SearchTreeSolver.ALMOST_FIXED_SCORE_VALUE)
		{
			this.almostFixedConflicts++;
		} else
		{
			this.normalConflicts++;
		}
		return greatestScoreConflict;
	}

	public int getLpSuccess()
	{
		return this.lpSuccess;
	}

	public long getNumberOfRecursiveSteps()
	{
		return this.iter;
	}

	@Override
	public int hashCode()
	{
		final int PRIM = 31;
		int out = this.hashCode();
		out *= PRIM;
		out += this.getParameterValue() * PRIM;
		out += this.modStatistic.hashCode() * PRIM;
		out += this.getConflictStorage().getConflicts().size() * PRIM;
		return out;
	}

	private void printValue(final int i, final int j, final int c, final int[] conflict)
	{
		for (int in = 0; in < this.level; in++)
		{
			System.out.print("\t");
		}
		System.out.println(this.level + "_d(" + i + "," + j + ")+" + c + ":k=" + this.getParameterValue() + ":#conflicts="
				+ this.getConflictStorage().getConflicts().size() + "<>" + new Conflicts(this).getConflicts().size() + "_"
				+ Arrays.toString(conflict));
	}

	/**
	 * Solves the given Hierarchical Clustering instance by a search tree algorithm.
	 * 
	 * @param kernelization true=Perform a two phase attack. At first compute a O(k^2)-node kernel and in the second step run the search tree
	 *            algorithm on this kernel. false= Runs the search tree algorithm on the whole instance.
	 * @return
	 * @throws UnsolvableException
	 */
	@Override
	public boolean solve()
	{
		this.iter++;

		// cannot find any conflict?
		if (this.getConflictStorage().isEmpty() && (this.getParameterValue() >= 0))
			return true;
		else if (this.getParameterValue() <= 0)
			return false;

		// interleaving?
		ChangeLogger logger = null;
		final int interleaveLevelDR = SearchTreeSolver.useLevelForKernelInterleavingConstant.getValue() ? this.level : this.getParameterValue();
		if (this.interleaveDR
				&& (((interleaveLevelDR + SearchTreeSolver.offset_DR.getValue()) % (interleaveLevelDR < SearchTreeSolver.interleaveDR_Bound_Large
						.getValue() ? SearchTreeSolver.interleaveDR_Constant_Small.getValue() : SearchTreeSolver.interleaveDR_Constant_Large
						.getValue())) == 0))
		{
			try
			{
				logger = this.performDataReductionRules(true);
			} catch (final UnsolvableException e)
			{
				return false;
			}

			// cannot find any conflict?
			if (this.getConflictStorage().isEmpty() && (this.getParameterValue() >= 0))
				return true;
			else if (this.getParameterValue() <= 0)
			{
				if (logger != null)
				{
					logger.reverseAllChanges();
				}
				return false;
			}
		}

		{ // lp
			final int interleaveLevelLP = SearchTreeSolver.useLevelForLPInterleavingConstant.getValue() ? this.level : this.getParameterValue();
			if (this.interleaveLP
					&& (((interleaveLevelLP + SearchTreeSolver.offset_LP.getValue()) % (interleaveLevelLP < SearchTreeSolver.interleaveLP_Bound_Small ? SearchTreeSolver.interleaveLP_Constant_Small
							.getValue()
							: interleaveLevelLP < SearchTreeSolver.interleaveLP_Bound_Medium.intValue() ? SearchTreeSolver.interleaveLP_Constant_Medium
									.getValue() : SearchTreeSolver.interleaveLP_Constant_Large.getValue())) == 0)
					&& (this.getParameterValue() > this.getNoOfElements()))
			{
				final HierClusterLpSolver lpSolver;

				if (this.lpHistory == null)
				{
					lpSolver = HierClusterLpSolver.getStandardLpSolver(this, -1, this.modStatistic);
				} else
				{
					final long key = SymmetricDiscreteDistanceFunction.hashSumDist(this) * this.modStatistic.getContentHash();

					if (this.lpHistory.containsKey(key))
					{
						// System.out.println("wiedergefunden" + key);
						lpSolver = this.lpHistory.get(key);// HierClusterLpSolver.getStandardLpSolver(this,
						// -1, this.modStatistic);
					} else
					{
						lpSolver = HierClusterLpSolver.getStandardLpSolver(this, -1, this.modStatistic);
						this.lpHistory.put(key, lpSolver);
					}
				}

				boolean exception = false;
				try
				{
					lpSolver.solve();
				} catch (final UnsolvableException e)
				{
					exception = true;
				}

				if (!exception)
				{
					if (lpSolver.isSolGreaterThan(this.getParameterValue()))
					{
						this.lpSuccess++;
						if (logger != null)
						{
							logger.reverseAllChanges();
						}
						return false;
					} else if (lpSolver.isIlpSolution())
					{
						this.set(lpSolver.getLpDistFunction());
						return true;
					}
				}
			}
		}

		// choose conflict
		final int[] conflict = this.chooseConflict();

		// branch into all 3 subcases
		final int i = conflict[0];
		final int j = conflict[1];
		final int l = conflict[2];
		final int d_i_j = this.get(i, j);
		final int d_j_l = this.get(j, l);
		final int d_i_l = this.get(i, l);

		if (this.modStatistic.isDecreasePossible(i, j))
		{
			this.changeDistanceValue(i, j, -1, true);
			if (this.verbose)
			{
				this.printValue(i, j, -1, conflict);
			}

			this.level++;
			if (!this.solve())
			{
				this.changeDistanceValue(i, j, 1, false);
				this.level--;
			} else
				return true;

			this.modStatistic.avoidDecrease(i, j);
		}

		if (this.modStatistic.isIncreasePossible(j, l))
		{
			final int c = Math.max(d_i_j, d_i_l) - d_j_l;

			this.changeDistanceValue(j, l, c, true);
			if (this.verbose)
			{
				this.printValue(j, l, c, conflict);
			}

			this.level++;
			final boolean result = this.solve();
			if (!result)
			{
				this.changeDistanceValue(j, l, -c, false);
				this.level--;
			} else
				return true;

			if (c == 1)
			{
				this.modStatistic.avoidIncrease(j, l);
			}
		}

		if (this.modStatistic.isIncreasePossible(i, l))
		{
			final int c = Math.max(d_j_l, d_i_j) - d_i_l;

			this.changeDistanceValue(i, l, c, true);
			if (this.verbose)
			{
				this.printValue(i, l, c, conflict);
			}

			this.level++;
			final boolean result = this.solve();
			if (!result)
			{
				this.changeDistanceValue(i, l, -c, false);
				this.level--;
			} else
				return true;
		}
		this.modStatistic.removeDecreaseBlock(i, j);
		this.modStatistic.removeIncreaseBlock(j, l);

		// restore status before kernelization
		if (logger != null)
		{
			logger.reverseAllChanges();
		}
		return false;
	}

	public boolean solve(final boolean kernelization, final boolean interleaving, final boolean interleaveLP)
	{
		if (kernelization)
		{
			if (SearchTreeSolver.SearchTreeSolver_LogLevel > 0)
			{
				System.out.println("start ini Rules");
			}
			try
			{
				this.performDataReductionRules(false);
			} catch (final UnsolvableException e)
			{
				if (SearchTreeSolver.SearchTreeSolver_LogLevel > 0)
				{
					System.out.println("Not solvable (by ini Data Reduction Rules)");
				}
				return false;
			}
		}
		if (SearchTreeSolver.SearchTreeSolver_LogLevel > 0)
		{
			System.out.println("end ini Rules");
		}

		this.interleaveDR = interleaving;
		this.interleaveLP = interleaveLP;
		return this.solve();
	}

	public boolean solve(final boolean kernelization, final boolean interleaving, final boolean interleaveLP,
			final TreeMap<Long, HierClusterLpSolver> lpHistory)
	{
		this.lpHistory = lpHistory;
		return this.solve(kernelization, interleaving, interleaveLP);
	}
}