package solver.lpBased;

import gurobi.GRB;
import gurobi.GRB.DoubleAttr;
import gurobi.GRB.DoubleParam;
import gurobi.GRBEnv;
import gurobi.GRBException;
import gurobi.GRBLinExpr;
import gurobi.GRBModel;
import gurobi.GRBVar;

import java.util.ArrayList;

import solver.hierarchicalClustering.ModificationStatistic;
import ProgOpt.SmacClusterEditOpt;
import dataReductionRules.UnsolvableException;
import distanceFunction.SymmetricDiscreteDistanceFunction;
import distanceFunction.SymmetricDistanceFunctionIntFunction;

public class Gurobi_LpSolver extends HierClusterLpSolver
{
	static GRBEnv env;

	static
	{
		try
		{
			Gurobi_LpSolver.env = new GRBEnv(null);
			Gurobi_LpSolver.env.set(GRB.IntParam.OutputFlag, 0);
			// Gurobi_LpSolver.env.set(DoubleParam.TimeLimit, 300);
		} catch (final GRBException e)
		{
			e.printStackTrace();
		}
	}

	/**
	 * Avoids resolving the same LP again and again, whenever you change the model set this flag to false
	 */
	protected boolean alreadySolved = false;

	// private boolean solveILP = false;

	private double ilpProbability = -0.1d;

	private final double maxTimeInSeconds;

	private ArrayList<SymmetricDistanceFunctionIntFunction<GRBVar>> variable;

	public Gurobi_LpSolver(final SymmetricDiscreteDistanceFunction dist, final int parameter, final ModificationStatistic modStatistic)
	{
		this(dist, parameter, modStatistic, Double.MAX_VALUE);
	}

	public Gurobi_LpSolver(final SymmetricDiscreteDistanceFunction dist, final int parameter, final ModificationStatistic modStatistic,
			final double maxTimeInSeconds)
	{
		super(dist, parameter, modStatistic);
		this.maxTimeInSeconds = maxTimeInSeconds;
	}

	private GRBModel getModel() throws GRBException
	{
		final GRBModel model = new GRBModel(Gurobi_LpSolver.env);

		final int n = this.getNoOfElements();

		// create variables
		this.variable = new ArrayList<SymmetricDistanceFunctionIntFunction<GRBVar>>();

		// GRBVar variables[][][] = new GRBVar[this.getMaxValue() - 1][][];
		{
			int index = 1;
			for (int t = 2; t <= this.getMaxValue(); t++)
			{
				this.variable.add(new SymmetricDistanceFunctionIntFunction<GRBVar>(n));
				// variable.set(t - 2, new
				// SymmetricDistanceFunction<GRBVar>(n));
				for (int i = 0; i < n; i++)
				{
					for (int j = i + 1; j < n; j++)
					{
						final String name = "x_t" + t + "_i" + i + "_j" + j;
						final int optCoeff = t <= this.get(i, j) ? -1 : 1;

						if ((this.modStatistic != null) && (this.get(i, j) >= t) && !this.modStatistic.isDecreasePossible(i, j))
						{
							this.variable.get(t - 2).set(i, j,
									model.addVar(1.0, 1.0, optCoeff, (Math.random() <= this.ilpProbability) ? GRB.INTEGER : GRB.CONTINUOUS, name));
						} else if ((this.modStatistic != null) && (t > this.get(i, j)) && !this.modStatistic.isIncreasePossible(i, j))
						{
							this.variable.get(t - 2).set(i, j,
									model.addVar(0.0, 0.0, optCoeff, (Math.random() <= this.ilpProbability) ? GRB.INTEGER : GRB.CONTINUOUS, name));
						} else
						{
							this.variable.get(t - 2).set(i, j,
									model.addVar(0.0, 1.0, optCoeff, (Math.random() <= this.ilpProbability) ? GRB.INTEGER : GRB.CONTINUOUS, name));
						}
						index++;
					}
				}
			}
			// Integrate new variables
			model.update();
		}

		{// model constraints

			for (int t = 2; t < this.getMaxValue(); t++)
			{
				for (int i = 0; i < n; i++)
				{
					for (int j = i + 1; j < n; j++)
					{
						final GRBLinExpr expr = new GRBLinExpr();
						expr.addTerm(1.0, this.variable.get(t - 2).get(i, j));
						expr.addTerm(-1.0, this.variable.get(t - 1).get(i, j));
						model.addConstr(expr, GRB.GREATER_EQUAL, 0.0, null);
					}
				}
			}

			for (int t = 2; t <= this.getMaxValue(); t++)
			{
				for (int i = 0; i < n; i++)
				{
					for (int j = i + 1; j < n; j++)
					{
						for (int l = j + 1; l < n; l++)
						{
							final GRBVar x_i_j_t = this.variable.get(t - 2).get(i, j);
							final GRBVar x_i_l_t = this.variable.get(t - 2).get(i, l);
							final GRBVar x_j_l_t = this.variable.get(t - 2).get(j, l);

							final GRBLinExpr exp1 = new GRBLinExpr();
							exp1.addTerm(1, x_i_j_t);
							exp1.addTerm(1, x_i_l_t);
							exp1.addTerm(-1, x_j_l_t);
							model.addConstr(exp1, GRB.GREATER_EQUAL, 0.0, null);

							final GRBLinExpr exp2 = new GRBLinExpr();
							exp2.addTerm(1, x_i_j_t);
							exp2.addTerm(-1, x_i_l_t);
							exp2.addTerm(1, x_j_l_t);
							model.addConstr(exp2, GRB.GREATER_EQUAL, 0.0, null);

							final GRBLinExpr exp3 = new GRBLinExpr();
							exp3.addTerm(-1, x_i_j_t);
							exp3.addTerm(1, x_i_l_t);
							exp3.addTerm(1, x_j_l_t);
							model.addConstr(exp3, GRB.GREATER_EQUAL, 0.0, null);
						}
					}
				}
			}
		}
		return model;

	}

	private boolean iniDistanceFunction(final ArrayList<SymmetricDistanceFunctionIntFunction<GRBVar>> variable, final int n) throws GRBException
	{
		boolean isIlpSolution = true;
		this.distFunc = new SymmetricDiscreteDistanceFunction(n);
		for (int i = 0; i < n; i++)
		{
			for (int j = i + 1; j < n; j++)
			{
				int d_i_j = 1;
				for (int t = 2; t <= this.getMaxValue(); t++)
				{

					final double value = variable.get(t - 2).get(i, j).get(DoubleAttr.X);

					if (!(Math.abs(value) <= 0.001d) && !(Math.abs(value - 1.0) <= 0.001))
					{
						isIlpSolution = false;
					}

					if (value > 0.5)
					{
						d_i_j++;
					} else
					{
						break;
					}
				}
				this.distFunc.set(i, j, d_i_j);
			}
		}
		return isIlpSolution;
	}

	@Override
	public void setILP(final boolean b)
	{
		this.setILPFraction(1d);
	}

	public void setILPFraction(final double ilpFraction)
	{
		if (ilpFraction > 1d)
			throw new IllegalArgumentException();

		this.ilpProbability = ilpFraction;
		this.alreadySolved = false;
	}

	@Override
	public boolean solve() throws UnsolvableException
	{
		if (!this.alreadySolved)
		{
			try
			// set max time
			{
				Gurobi_LpSolver.env.set(DoubleParam.TimeLimit, Math.min(this.maxTimeInSeconds, SmacClusterEditOpt.cutoff_time.doubleValue()));
			} catch (final GRBException e1)
			{
				e1.printStackTrace();
			}
			try
			{
				final GRBModel model = this.getModel();
				model.getEnv().set(GRB.DoubleParam.TimeLimit, Math.min(this.maxTimeInSeconds, SmacClusterEditOpt.cutoff_time.doubleValue()));
				model.optimize();

				final int status = model.get(GRB.IntAttr.Status);
				if (status == GRB.OPTIMAL)
				{
					this.lpSolutionVal = (float) model.get(GRB.DoubleAttr.ObjVal) + SymmetricDiscreteDistanceFunction.sumDist(this);
					this.isIlpSolution = this.iniDistanceFunction(this.variable, this.getNoOfElements());
				} else
					// System.err.println("abort lp but lB:"+model.get(DoubleAttr.ObjVal));
					throw new UnsolvableException("LP cannot be solved optimal.");

			} catch (final GRBException e)
			{
				throw new UnsolvableException();
			}
		}
		this.alreadySolved = true;
		return this.isIlpSolution;
	}
}
