package solver.lpBased;

import org.apache.commons.lang3.mutable.MutableInt;

import solver.hierarchicalClustering.HierarchicalClusteringSolver;
import solver.hierarchicalClustering.ModificationStatistic;
import benchmark.helpClasses.StopWatch;
import dataReductionRules.UnsolvableException;
import distanceFunction.SymmetricDiscreteDistanceFunction;

public abstract class HierClusterLpSolver extends HierarchicalClusteringSolver
{
	public static final MutableInt LP_LowerBound_ILP_Part_First_Run = new MutableInt(5);

	public static final boolean verbose = false;

	public static int getLpLowerBound(final SymmetricDiscreteDistanceFunction dist) throws FoundILPSolution
	{
		return HierClusterLpSolver.getLpLowerBound(dist, -1, -1d);
	}

	/**
	 * Returns a lower bound on the parameter value.
	 * 
	 * @param maxTime
	 * @throws FoundILPSolution
	 */
	public static int getLpLowerBound(final SymmetricDiscreteDistanceFunction dist, final double maxTime, final double ilpStartFrac)
			throws FoundILPSolution
	{
		final StopWatch watch = new StopWatch();

		int lB = 0;
		double ilpFraction = ilpStartFrac < 0d ? -1d : HierClusterLpSolver.minFrac(dist);
		while (watch.stopp().getTimeInSec() <= maxTime)
		{
			if (HierClusterLpSolver.verbose)
			{
				System.err.println("run lp lowerbound with ilp Frac:" + ilpFraction + " [current lower bound:" + lB + ", remain time:"
						+ (maxTime - watch.stopp().getTimeInSec()) + "]");
			}

			final Gurobi_LpSolver partILPSolver = new Gurobi_LpSolver(dist, -1, null, Math.max(1, maxTime - watch.stopp().getTimeInSec()));
			partILPSolver.setILPFraction(ilpFraction);

			try
			{
				partILPSolver.solve();
				lB = Math.max(lB, (int) Math.ceil(partILPSolver.getOptLpSolution()));
			} catch (final UnsolvableException e)
			{
			}

			if (partILPSolver.isIlpSolution)
			{
				if (HierClusterLpSolver.verbose)
				{
					System.err.println("found ILP solution in LP lower bound run:" + ilpFraction + " (" + watch.stopp().getTimeInSec() + "s)");
				}
				throw new FoundILPSolution(lB, partILPSolver);
			}

			ilpFraction = ilpFraction < 0d ? HierClusterLpSolver.minFrac(dist) : (ilpFraction * 2d);
			ilpFraction = Math.min(ilpFraction, 1d);
		}
		return lB;
	}

	public static HierClusterLpSolver getStandardLpSolver(final SymmetricDiscreteDistanceFunction searchTreeSolver, final int param)
	{
		return HierClusterLpSolver.getStandardLpSolver(searchTreeSolver, param, null);
	}

	public static HierClusterLpSolver getStandardLpSolver(final SymmetricDiscreteDistanceFunction searchTreeSolver, final int param,
			final ModificationStatistic mod)
	{
		// return new GLPK_LpSolver(searchTreeSolver, param,mod);
		return new Gurobi_LpSolver(searchTreeSolver, param, mod);
	};

	public static double minFrac(final SymmetricDiscreteDistanceFunction dist)
	{
		return Math.min(1d, Math.max(HierClusterLpSolver.LP_LowerBound_ILP_Part_First_Run.doubleValue() / dist.getNoOfElements(), 0.01d));
	}

	/**
	 * Distance function provided by the lp solution.
	 */
	protected SymmetricDiscreteDistanceFunction distFunc;

	protected boolean isIlpSolution = false;

	protected float lpSolutionVal = Float.MAX_VALUE;

	public HierClusterLpSolver(final SymmetricDiscreteDistanceFunction dist, final int parameter, final ModificationStatistic modStatistic)
	{
		super(dist, parameter);
		this.modStatistic = modStatistic;
	}

	public SymmetricDiscreteDistanceFunction getLpDistFunction()
	{
		return this.distFunc;
	}

	public double getOptLpSolution()
	{
		return this.lpSolutionVal;
	}

	public boolean isIlpSolution()
	{
		return this.isIlpSolution;
	}

	/**
	 * Return true, if the lower bound provided by the LP solution is greater than the parameter.
	 * 
	 * @param parameterValue
	 * @return
	 * @throws UnsolvableException
	 */
	public boolean isSolGreaterThan(final int parameterValue)
	{
		return (int) Math.ceil(this.getOptLpSolution()) > parameterValue;
	}

	public abstract void setILP(boolean b);

	@Override
	public abstract boolean solve() throws UnsolvableException;
}